import re
from datetime import datetime, timedelta
from functools import total_ordering

import pandas as pd
import tzlocal
from icalendar import Calendar, Event
from pytz import timezone
import numpy as np

@total_ordering
class OfficeHour:
    def __init__(self, name):
        self.day_idx = normalize_day_of_week(name)
        time_start, time_end = name.split('-')
        self.time_start = to_time(time_start)
        self.time_end = to_time(time_end)
        self.name = name

    def to_tuple(self):
        return self.day_idx, self.time_start, self.time_end

    def __str__(self):
        return f'OfficeHour({self.name})'

    def __lt__(self, other):
        return self.to_tuple() < other.to_tuple()

    def __eq__(self, other):
        return self.to_tuple() == other.to_tuple()

    def intersects(self, other):
        if self.day_idx != other.day_idx:
            # different days, can't intersect
            return False

        if self < other:
            # self begins first (or at same time)
            return self.time_end > other.time_start
        else:
            # other begins first (or at same time)
            return other.time_end > self.time_start

    def get_event_kwargs(self, date_start, date_end, tz=None, **kwargs):
        """ gets weekly recurring event arguments, to be passed to Event

        Args:
            date_start (str): start date
            date_end (str): end date
            tz (str or timezone, optional): timezone. If not given, the local
                timezone is used.
            **kwargs: Additional parameters to be included in the event

        Returns:
            kwargs: dictionary to be unapcked into Event object

        Raises:
            AttributeError: event exceeds the maximum weekly repeats (53 weeks)
        """
        # Convert date_start, date_end to date objects
        date_start = pd.to_datetime(date_start).date()
        date_end = pd.to_datetime(date_end).date()

        # Move start date up to the correct weekday
        while date_start.weekday() != self.day_idx:
            date_start += timedelta(days=1)

        # Handle timezone (default to local time if not provided)
        if tz is None:
            tz = tzlocal.get_localzone()
        tz = timezone(str(tz))

        # build output starts / ends
        dtstart = datetime.combine(date_start, self.time_start)
        kwargs['dtstart'] = tz.localize(dtstart)
        dtend = datetime.combine(date_start, self.time_end)
        kwargs['dtend'] = tz.localize(dtend)

        # Compute the number of weekly repeats before the end date
        date = date_start
        for repeats in range(52):
            if date > date_end:
                break
            date = date + timedelta(weeks=1)
        else:
            raise AttributeError(f'> 1 yr event: {date_start} to {date_end}')

        kwargs['rrule'] = {'freq': 'weekly', 'count': repeats}

        return kwargs


def get_intersection_dict(oh_list):
    # order office hours (from earliest to latest in the day)
    _oh_list = [OfficeHour(oh) for oh in oh_list]
    idx_map = np.argsort(_oh_list)
    _oh_list.sort()

    # find intersections (initialize with reflexivity)
    oh_int_dict = {idx: [idx] for idx in range(len(oh_list))}
    for idx0, oh0 in enumerate(_oh_list):
        for _idx1, oh1 in enumerate(_oh_list[idx0 + 1:]):
            # idx1 is consistent with ordering of _oh_list
            idx1 = _idx1 + idx0 + 1
            if oh0.intersects(oh1):
                oh_int_dict[idx0].append(idx1)
                oh_int_dict[idx1].append(idx0)
            else:
                # oh0 is before oh1, if oh0 doesn't intersect oh1 it can't
                # intersect any which come after it in _oh_list, its sorted
                break

    # map from indexing of _oh_list back to given oh_list indexing
    oh_int_dict = {idx_map[k]: [idx_map[_v] for _v in v]
                   for k, v in oh_int_dict.items()}

    return oh_int_dict


def normalize_day_of_week(day_str):
    """ extracts a day of week, as index, from a string

    Args:
        day_str (str): string containing some day of the week

    Returns:
        day_idx (int): 0 for monday, 1 for tuesday, ...
    """
    date_regex = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']
    match_list = [bool(re.search(pattern, day_str, re.IGNORECASE))
                  for pattern in date_regex]

    assert sum(match_list) < 2, f'non-unique day of week found in: {day_str}'
    assert sum(match_list) == 1, f'no day of week found in: {day_str}'

    # return idx of first match in list
    for idx, b in enumerate(match_list):
        if b:
            return idx
    return None


def to_time(time_str):
    """ converts to timedelta, from beginning of day to time_str

    Args:
        time_str (str): comes in one of two formats: "6:30 PM" or "4 AM"

    Returns:
        delta (timedelta): from beginning of day
    """
    # Match patterns for 12-hour and 24-hour unambiguous formats
    patterns = [
        ('%I:%M%p', re.compile(r'\d{1,2}:\d{2}\s*(?:AM|PM)', re.IGNORECASE)),
        ('%I%p', re.compile(r'\d{1,2}\s*(?:AM|PM)', re.IGNORECASE))
    ]

    for fmt, pattern in patterns:
        match_list = pattern.findall(time_str)
        match len(match_list):
            case 0:
                # no match found
                continue
            case 1:
                # unique match found
                s_match = match_list[0].replace(' ', '')
                return datetime.strptime(s_match, fmt).time()
            case _:
                raise ValueError(f'Multiple times found: {time_str}')

    raise ValueError(f'Ambiguous or invalid time string: "{time_str}"')


def build_calendar(oh_ta_dict, date_start, date_end, **kwargs):
    """  builds a calendar, a set of events, from oh_ta_dict

    Args:
        oh_ta_dict (dict): keys are office hours slots (see time_str in
            get_event_kwargs()), values are lists of str (TA names)
        date_start (str): starting date for office hours for course
            (inclusive), see  get_event_kwargs()
        date_end (str): ending date for office hours for course (inclusive),
            see get_event_kwargs()=

    Returns:
        cal (Calendar): ready to be exported to ics format
    """
    cal = Calendar()
    for oh_name, ta_list in oh_ta_dict.items():
        if not ta_list:
            # skip oh slots without any TAs
            continue
        ta_list = [ta.capitalize() for ta in sorted(ta_list)]
        summary = ', '.join(sorted(ta_list))

        oh = OfficeHour(name=oh_name)
        _kwargs = oh.get_event_kwargs(summary=summary,
                                     date_start=date_start,
                                     date_end=date_end,
                                     **kwargs)

        # build event with proper attributes of event,
        _kwargs = kwargs | _kwargs
        cal.add_component(Event(**_kwargs))

    return cal



import numpy as np

import oh_sched
from oh_sched.config import Config
from oh_sched.email import find_similar_str


def main(f_csv, config):
    prefs, email_list, name_list, oh_list = oh_sched.extract_csv(f_csv)

    if config.verbose:
        # print message about availability given by TAs
        num_available = prefs.shape[1] - np.isnan(prefs).sum(axis=1)
        print(f'TAs with lowest 20% of availability given:')
        for ta_idx in np.argsort(num_available)[:len(email_list) // 5]:
            n = num_available[ta_idx]
            print(f'  {n} OH slots possible: {email_list[ta_idx]}')

    # scale per day
    if config.scale_dict is not None:
        prefs = prefs * oh_sched.get_scale(oh_list,
                                           scale_dict=config.scale_dict,
                                           verbose=config.verbose)

    # match
    oh_ta_match = oh_sched.match(prefs,
                                 oh_per_ta=config.oh_per_ta,
                                 max_ta_per_oh=config.max_ta_per_oh)

    perc_max = oh_sched.get_perc_max(oh_ta_match, prefs=prefs)
    assert not np.isnan(perc_max).any(), 'TA assigned outside availability'

    # export to ics
    oh_ta_dict = {oh_list[oh]: [name_list[ta] for ta in ta_list]
                  for oh, ta_list in enumerate(oh_ta_match)}
    config.to_ics(oh_ta_dict)

    if config.verbose:
        # print TAs per slot
        print('\nSchedule:')
        for oh, ta_list in oh_ta_dict.items():
            if not len(ta_list):
                continue

            ta_csv = ', '.join(ta_list)
            print(f'{oh} has {len(ta_list)} TAs: {ta_csv}')

        print('\nPercentage Max Score :')
        print(
            'https://github.com/matthigger/oh_sched?tab=readme-ov-file#percentage-max')
        print(f'min percentage max score: {perc_max.min():.4f}')
        print(f'mean percentage max score: {perc_max.mean():.4f}\n')

    # warn on similar emails
    email_tup_list = list(find_similar_str(email_list, max_distance=2))
    if email_tup_list:
        print('WARNING: similar emails treated as unique')
        print('https://github.com/matthigger/oh_sched?tab=readme-ov-file#email-comparison')
        for email0, email1 in email_tup_list:
            print(f'{email0} vs \n{email1}\n')
        print('')


if __name__ == '__main__':
    import argparse
    import pathlib

    parser = argparse.ArgumentParser(
        description='https://github.com/matthigger/oh_sched')
    parser.add_argument('f_csv', type=str,
                        help='Path to the TA OH preference CSV')
    parser.add_argument('-c', '--config', type=str, default=None,
                        help='path to yaml file with configuration.  you may use (and create) a default configuration yaml by not passing this parameter')
    param = parser.parse_args()

    if param.config is None:
        f_config = pathlib.Path('config.yaml')

        if f_config.exists():
            print(f'Using config file {f_config}, pass with -c {f_config} to '
                  f'avoid this message')
            config = Config.from_yaml(f_config)
        else:
            # build default config, dump to file for user
            print(f'Using default config {f_config} please revise as needed '
                  f'(https://github.com/matthigger/oh_sched)')
            config = Config()
            config.to_yaml(f_config)
    else:
        config = Config.from_yaml(param.config)

    main(param.f_csv, config)



import datetime
import pathlib
import re
import warnings
from copy import copy
from datetime import datetime, timedelta

import pytz
import yaml

import oh_sched.calendr

OH_ALL = 'ALL'
OH_LEFTOVER = 'LEFTOVER'


class Config:
    def __init__(self, oh_per_ta=1, max_ta_per_oh=None, scale_dict=None,
                 date_start=None, date_end=None, f_out=None, tz=None,
                 f_out_dict=None,
                 verbose=True):

        self.oh_per_ta = int(oh_per_ta)
        assert self.oh_per_ta > 0

        self.max_ta_per_oh = max_ta_per_oh
        if self.max_ta_per_oh is not None:
            self.max_ta_per_oh = int(max_ta_per_oh)
            assert self.max_ta_per_oh > 0

        self.scale_dict = scale_dict
        if self.scale_dict is not None:
            self.scale_dict = {str(k): float(v) for k, v in scale_dict.items()}

        today = datetime.today()
        if date_start is None:
            date_start = today.strftime("%b %d %Y")
        self.date_start = str(date_start)

        if date_end is None:
            date_end = (today + timedelta(weeks=1)).strftime(
                "%b %d %Y")
        self.date_end = str(date_end)

        assert (f_out is None) or (f_out_dict is None), \
            'cannot pass both f_out and f_out_dict'

        if f_out is None:
            f_out = 'oh.ics'
        if f_out_dict is None:
            f_out_dict = {f_out: OH_ALL}
        self.f_out_dict = {pathlib.Path(f_out): v
                           for f_out, v in f_out_dict.items()}
        for f_out in self.f_out_dict.keys():
            assert f_out.parent.exists()

        self.tz = tz
        if self.tz is None:
            self.tz = 'US/Eastern'
        else:
            if self.tz not in pytz.all_timezones:
                raise AttributeError(f'timezone not found in IANA database')

        self.verbose = bool(verbose)

    @classmethod
    def from_yaml(cls, f_yaml):
        with open(str(f_yaml), 'r') as f:
            config = yaml.safe_load(f)
        return Config(**config)

    def to_yaml(self, f_yaml):
        with open(str(f_yaml), 'w') as f:
            yaml.dump(self.to_dict(), f)

    def to_dict(self):
        d = copy(self.__dict__)
        if d['f_out'] is not None:
            d['f_out'] = str(d['f_out'])
        return d

    def to_ics(self, oh_ta_dict, verbose=True):
        """ writes schedule to ics files

        Args:
            oh_ta_dict (dict): keys are office hour slots (str), values are
                list of str, names of TAs assigned to this slot

        Returns:
            f_dict (dict): keys are file names (from self.f_out_dict.keys()),
                values are oh_ta_dict (of form above) which are relevant for
                corresponding file
        """
        # build f_dict, whose keys are output files and whose strings are the
        # corresponding oh_ta_dict to be written in each
        oh_leftover = set(oh_ta_dict.keys())
        f_leftover = list()
        f_dict = dict()
        for f_out, s_regex in self.f_out_dict.items():
            if s_regex == OH_ALL:
                # all office hours slots in this output
                _oh_ta_dict = oh_ta_dict
                oh_leftover = set()
            elif s_regex == OH_LEFTOVER:
                # leftover office hours slots in this output (handle later)
                f_leftover.append(f_out)
                continue
            else:
                # search for s_regex, include only oh slots which match
                regex = re.compile(s_regex)
                _oh_ta_dict = dict()
                for oh, ta_list in oh_ta_dict.items():
                    if regex.search(oh):
                        # add this office hours to this calendar
                        _oh_ta_dict[oh] = ta_list
                        # remove from leftover
                        oh_leftover.discard(oh)

            f_dict[f_out] = _oh_ta_dict

        # add leftover oh to their corresponding files
        if oh_leftover:
            _oh_ta_dict = {oh: oh_ta_dict[oh] for oh in oh_leftover}
            for f_out in f_leftover:
                f_dict[f_out] = _oh_ta_dict
            if not f_leftover:
                s = sorted(oh_leftover)
                warnings.warn(f'OH not written to any ics outputs: {s}')

        # write ics files
        for f_out, _oh_ta_dict in f_dict.items():
            cal = oh_sched.calendr.build_calendar(_oh_ta_dict,
                                                  date_start=self.date_start,
                                                  date_end=self.date_end,
                                                  tz=self.tz)

            with open(f_out, 'wb') as f:
                f.write(cal.to_ical())

            if verbose:
                print(f'writing calendar file: {f_out}\n')

        return f_dict


if __name__ == '__main__':
    import oh_sched

    # dump default config to test directory
    folder = pathlib.Path(oh_sched.__file__).parents[1] / 'test'
    config = Config()
    config.to_yaml(folder / 'config.yaml')



import Levenshtein


def find_similar_str(s_list, max_distance=2, case_insensitive=True):
    """ returns iterator of strings with edit distance of 3 (or less) apart

    used here to find email typos

    Args:
        s_list (list): list of strings to compare
        max_distance (int): maximum levenshtein distance at which two
            strings are still declared "similar"

    Yields:
        s_tup (tuple): (2) strings which are deemed "similar"
    """
    if case_insensitive:
        s_list = [s.lower() for s in s_list]

    for idx, s0 in enumerate(s_list):
        for s1 in s_list[idx + 1:]:
            if Levenshtein.distance(s0, s1) <= max_distance:
                yield s0, s1
